use pest::{iterators::{Pair, Pairs}, Parser};
use crate::parser::{DeimosParser, Rule};

#[derive(Debug)]
pub struct CompilationUnit {
    namespace: Option<Fqn>,
    declarations: Vec<Declaration>,
}

#[derive(Debug)]
pub struct Fqn {
    identifiers: Vec<Identifier>
}

#[derive(Debug)]
pub enum Declaration {
    Interface {
        identifier: Identifier,
        is_extern: bool,
        is_public: bool,
        type_declaration: TypeDeclaration,
    },
    Implementation {
        identifier: Identifier,
        is_extern: bool,
        is_public: bool,
        type_declaration: TypeDeclaration
    },
    Module {
        identifier: Identifier,
        is_extern: bool,
        is_public: bool,
        declarations: Vec<Declaration>,
    },
    Function {
        identifier: Identifier,
        is_extern: bool,
        is_public: bool,
        statement: Statement 
    }
}

#[derive(Debug)]
pub struct TypeDeclaration {
    generics: Vec<GenericParameter>,
    extends: Vec<Type>,
    declarations: Vec<Declaration>,
}

impl TypeDeclaration {
    pub fn new(generics: Vec<GenericParameter>, extends: Vec<Type>, declarations: Vec<Declaration>) -> TypeDeclaration {
        TypeDeclaration {
            generics,
            extends,
            declarations
        }
    }
}

#[derive(Debug)]
pub struct Identifier {
    name: String
}

#[derive(Debug)]
pub struct GenericParameter {
    identifier: Identifier,
    bound: Option<GenericBound>
}

#[derive(Debug)]
pub enum GenericBound {
    Extends {
        fqn: Fqn
    },
    Super {
        fqn: Fqn
    }
}

#[derive(Debug)]
pub struct GenericArgument {
    fqn: Fqn
}

#[derive(Debug)]
pub struct Type {
    fqn: Fqn,
    generics: Vec<GenericArgument>
}

#[derive(Debug)]
pub enum Statement {
    BlockStatement {
        statements: Vec<Statement>
    },
    CallStatement,
    AssignmentStatement
}

fn build_generic_argument(generic_argument_pair: Pair<Rule>) -> GenericArgument {
    let fqn_pair = generic_argument_pair.into_inner().next().unwrap();
    GenericArgument {
        fqn: build_fqn(fqn_pair)
    }
}

fn build_generic_arguments_declaration(generic_arguments_declaration_pair: Pair<Rule>) -> Vec<GenericArgument> {
    let mut generic_arguments: Vec<GenericArgument> = vec![];
    for pair in generic_arguments_declaration_pair.into_inner() {
        match pair.as_rule() {
            Rule::generic_argument => {
                generic_arguments.push(build_generic_argument(pair));
            }
            _ => panic!("Expected only generic_argument rules. Found: {}", pair)
        }
    }
    generic_arguments
}

fn build_generic_parameter(generic_parameter_pair: Pair<Rule>) -> GenericParameter {
    let identifier_pair = generic_parameter_pair.into_inner().next().unwrap();
    GenericParameter {
        identifier: build_identifier(identifier_pair),
        bound: None,
    }
}

fn build_type(type_pair: Pair<Rule>) -> Type {
    let mut fqn: Option<Fqn> = None;
    let mut generic_arguments_declaration: Option<Vec<GenericArgument>> = None;

    for pair in type_pair.into_inner() {
        match pair.as_rule() {
            Rule::fqn => {
                fqn = Some(build_fqn(pair));
            }
            Rule::generic_arguments_declaration => {
                generic_arguments_declaration = Some(build_generic_arguments_declaration(pair));
            }
            _ => panic!("Expected only fqn or generic_arguments_declaration rules. Found: {}", pair)
        }
    }

    Type {
        fqn: fqn.unwrap(),
        generics: generic_arguments_declaration.unwrap_or(vec![])
    }
}

fn build_generic_parameters_declaration(generic_parameters_declaration_pair: Pair<Rule>) -> Vec<GenericParameter> {
    let mut parameters: Vec<GenericParameter> = vec![];
    for pair in generic_parameters_declaration_pair.into_inner() {
        match pair.as_rule() {
            Rule::generic_parameter => {
                parameters.push(build_generic_parameter(pair));
            }
            _ => panic!("Expected only generic_parameter rule. Found: {}", pair)
        }
    }
    parameters
}

fn build_extends_list(extends_list_pair: Pair<Rule>) -> Vec<Type> {
    let mut extensions: Vec<Type> = vec![];
    
    for pair in extends_list_pair.into_inner() {
        match pair.as_rule() {
            Rule::r#type => {
                extensions.push(build_type(pair));            
            }
            _ => panic!("Expected only type rule. Found: {}", pair)
        } 
    }

    extensions
}

fn build_interface(is_extern: bool, is_public: bool, interface_pair: Pair<Rule>) -> Declaration {
    let mut identifier: Option<Identifier> = None;
    let mut generic_parameters: Option<Vec<GenericParameter>> = None;
    let mut extends: Option<Vec<Type>> = None;
    let mut declarations: Vec<Declaration> = vec![];
   
    for pair in interface_pair.into_inner() {
        match pair.as_rule() {
            Rule::identifier => {
                identifier = Some(build_identifier(pair));
            }
            Rule::generic_parameters_declaration => {
                generic_parameters = Some(build_generic_parameters_declaration(pair));
            }
            Rule::extends_list => {
                extends = Some(build_extends_list(pair));    
            }
            Rule::declaration => {
                declarations.push(build_declaration(pair));
            },
            _ => panic!(
                "Expected only identifier, generics_declaration, extends_list, or declaration rules. Found: {}",
                pair
            )
        }
    }

    Declaration::Interface { 
        identifier: identifier.unwrap(), 
        is_extern, 
        is_public, 
        type_declaration: TypeDeclaration::new(
            generic_parameters.unwrap_or(vec![]),
            extends.unwrap_or(vec![]), 
            declarations
        )
    }
}

fn build_implementation(is_extern: bool, is_public: bool, pair: Pair<Rule>) -> Declaration {
    todo!()
}

fn build_module(is_extern: bool, is_public: bool, pair: Pair<Rule>) -> Declaration {
    todo!()
}

fn build_function(is_extern: bool, is_public: bool, pair: Pair<Rule>) -> Declaration {
    todo!()
}

fn build_declaration(declaration_pair: Pair<Rule>) -> Declaration {
    let mut is_extern = false;
    let mut is_public = false;
    let mut declaration: Option<Declaration> = None;

    for pair in declaration_pair.into_inner() {
        match pair.as_rule() {
            Rule::r#extern => {
                is_extern = true;
            },
            Rule::r#pub => {
                is_public = true;
            }
            Rule::interface => { 
                declaration = Some(build_interface(is_extern, is_public, pair));
            },
            Rule::implementation => {
                declaration = Some(build_implementation(is_extern, is_public, pair));
            },
            Rule::module => {
                declaration = Some(build_module(is_extern, is_public, pair));
            },
            Rule::function => {
                declaration = Some(build_function(is_extern, is_public, pair));
            },
            _ => panic!("Expected only interface, implementation, module, or function rules; found {}", pair)
        }
    }
    declaration.expect("Expected declaration.")
}

fn build_identifier(pair: Pair<Rule>) -> Identifier {
    match pair.as_rule() {
        Rule::identifier => {
            Identifier {
                name: String::from(pair.as_str())
            }
        }
        _ => panic!("Expected an identifier.") 
    } 
}

fn build_fqn(fqn_pair: Pair<Rule>) -> Fqn {
    let mut identifiers: Vec<Identifier> = vec![];
    for pair in fqn_pair.into_inner() {
        match pair.as_rule() {
            Rule::identifier => {
                identifiers.push(build_identifier(pair));
            }
            _ => panic!("Expected only identifiers.")
        }
    }
    Fqn {
        identifiers
    }
}

fn build_compilation_unit(pair: Pair<Rule>)-> CompilationUnit {
    let mut namespace: Option<Fqn> = None;
    let mut declarations: Vec<Declaration> = vec![];

    for pair in pair.into_inner() {
        match pair.as_rule() {
            Rule::namespace => {
                let fqn_pair = pair.into_inner().next().unwrap(); 
                namespace = Some(build_fqn(fqn_pair));
            }
            Rule::declaration => {
                declarations.push(build_declaration(pair));
            }
            Rule::EOI => {} // ignore
            _ => panic!("Expected only namespace, declaration, or EOI rules, found: {}", pair)
        }    
    }

    CompilationUnit {
        namespace,
        declarations
    }
}

pub fn build_ast(src: &str) -> CompilationUnit {
    let pair = DeimosParser::parse(Rule::compilation_unit, src)
        .expect("Unsuccessful parse.")
        .next()
        .expect("Expcted compilation_unit.");
    match pair.as_rule() {
        Rule::compilation_unit => {
            build_compilation_unit(pair)
        },
        _ => panic!("Expected compilation_unit rule.")
    }
}

